@@ -35,35 +35,29 @@
 import string
 import sys
 import re
+import os
 import traceback
 
-from gi.repository import Gtk, Gdk, GObject, Pango, Peas
+from gi.repository import GLib, Gtk, Gdk, Gio, GObject, Pango, Peas
 from gi.repository import RB
 
 import gettext
 gettext.install('rhythmbox', RB.locale_dir())
 
+DEBUG_PORT = 5678
+
 try:
 	import rpdb2
 	have_rpdb2 = True
 except:
 	have_rpdb2 = False
 
-import gettext
-gettext.install('rhythmbox', RB.locale_dir())
+try:
+	import debugpy
+	have_debugpy = True
+except:
+	have_debugpy = False
 
-ui_str = """
-<ui>
-  <menubar name="MenuBar">
-    <menu name="ToolsMenu" action="Tools">
-      <placeholder name="ToolsOps_5">
-        <menuitem name="PythonConsole" action="PythonConsole"/>
-        <menuitem name="PythonDebugger" action="PythonDebugger"/>
-      </placeholder>
-    </menu>
-  </menubar>
-</ui>
-"""
 
 class PythonConsolePlugin(GObject.Object, Peas.Activatable):
 	__gtype_name__ = 'PythonConsolePlugin'
@@ -73,50 +67,58 @@ class PythonConsolePlugin(GObject.Object, Peas.Activatable):
 	def __init__(self):
 		GObject.Object.__init__(self)
 		self.window = None
-			
+		
 	def do_activate(self):
-		data = dict()
 		shell = self.object
-		manager = shell.props.ui_manager
-		
-		data['action_group'] = Gtk.ActionGroup(name='PythonConsolePluginActions')
+		app = shell.props.application
 
-		action = Gtk.Action(name='PythonConsole', label=_("_Python Console"),
-		                    tooltip=_("Show Rhythmbox's python console"),
-		                    stock_id='gnome-mime-text-x-python')
+		action = Gio.SimpleAction.new("python-console", None)
 		action.connect('activate', self.show_console, shell)
-		data['action_group'].add_action(action)
+		app.add_action(action)
+
+		app.add_plugin_menu_item("tools",
+					 "python-console",
+					 Gio.MenuItem.new(label=_("Python Console"),
+						 	  detailed_action="app.python-console"))
+
+		self.rpdb2_action = Gio.SimpleAction.new("python-debugger-winpdb", None)
+		self.rpdb2_action.connect('activate', self.attach_winpdb, shell)
+		app.add_action(self.rpdb2_action)
+
+		app.add_plugin_menu_item("tools",
+					 "python-debugger-winpdb",
+					 Gio.MenuItem.new(label=_("Python Debugger (winpdb)"),
+							  detailed_action="app.python-debugger-winpdb"))
+
+		self.debugpy_action = Gio.SimpleAction.new("python-debugger-debugpy", None)
+		self.debugpy_action.connect('activate', self.enable_debugpy, shell)
+		app.add_action(self.debugpy_action)
+
+		app.add_plugin_menu_item("tools",
+					 "python-debugger-debugpy",
+					 Gio.MenuItem.new(label=_("Python Debugger (debugpy)"),
+							  detailed_action="app.python-debugger-debugpy"))
 
-		action = Gtk.Action(name='PythonDebugger', label=_("Python Debugger"),
-				    tooltip=_("Enable remote python debugging with rpdb2"),
-				    stock_id=None)
-		if have_rpdb2:
-			action.connect('activate', self.enable_debugging, shell)
-		else:
-			action.set_visible(False)
-		data['action_group'].add_action(action)
-				
-		manager.insert_action_group(data['action_group'], 0)
-		data['ui_id'] = manager.add_ui_from_string(ui_str)
-		manager.ensure_update()
-		
-		shell.PythonConsolePluginInfo = data
-	
 	def do_deactivate(self):
 		shell = self.object
-		data = shell.PythonConsolePluginInfo
+		app = shell.props.application
 
-		manager = shell.props.ui_manager
-		manager.remove_ui(data['ui_id'])
-		manager.remove_action_group(data['action_group'])
-		manager.ensure_update()
+		app.remove_plugin_menu_item("tools", "python-console")
+		app.remove_action("python-console")
 
-		shell.PythonConsolePluginInfo = None
+		app.remove_plugin_menu_item("tools", "python-debugger-winpdb")
+		app.remove_action("python-debugger-winpdb")
+		self.rpdb2_action = None
+
+		app.remove_plugin_menu_item("tools", "python-debugger-debugpy")
+		app.remove_action("python-debugger-debugpy")
+		self.debugpy_action = None
 		
 		if self.window is not None:
 			self.window.destroy()
-		
-	def show_console(self, action, shell):
+
+
+	def show_console(self, action, parameter, shell):
 		if not self.window:
 			ns = {'__builtins__' : __builtins__, 
 			      'RB' : RB,
@@ -124,11 +126,11 @@ class PythonConsolePlugin(GObject.Object, Peas.Activatable):
 			console = PythonConsole(namespace = ns, 
 			                        destroy_cb = self.destroy_console)
 			console.set_size_request(600, 400)
-			console.eval('print "' + \
+			console.eval('print("' + \
 			             _("You can access the main window " \
-			             "through the \'shell\' variable :") +  
-			             '\\n%s" % shell', False)
-	
+			             "through the \'shell\' variable :") +
+			             '\\n%s" % shell)', False)
+
 			self.window = Gtk.Window()
 			self.window.set_title('Rhythmbox Python Console')
 			self.window.add(console)
@@ -138,11 +140,16 @@ class PythonConsolePlugin(GObject.Object, Peas.Activatable):
 			self.window.show_all()
 		self.window.grab_focus()
 
-	def enable_debugging(self, action, shell):
-		pwd_path = os.path.join(rb.user_data_dir(), "rpdb2_password")
+	def attach_winpdb(self, action, parameter, shell):
+		if have_rpdb2 is False:
+			self.missing_python_module("rpdb2", "winpdb")
+			self.rpdb2_action.set_enabled(False)
+			return False
+
+		pwd_path = os.path.join(RB.user_data_dir(), "rpdb2_password")
 		msg = _("After you press OK, Rhythmbox will wait until you connect to it with winpdb or rpdb2. If you have not set a debugger password in the file %s, it will use the default password ('rhythmbox').") % pwd_path
 		dialog = Gtk.MessageDialog(None, 0, Gtk.MessageType.INFO, Gtk.ButtonsType.OK_CANCEL, msg)
-		if dialog.run() == Gtk.RESPONSE_OK:
+		if dialog.run() == Gtk.ResponseType.OK:
 			password = "rhythmbox"
 			if os.path.exists(pwd_path):
 				pwd_file = open(pwd_path)
@@ -153,9 +160,41 @@ class PythonConsolePlugin(GObject.Object, Peas.Activatable):
 				rpdb2.start_embedded_debugger(password)
 				return False
 
-			GObject.idle_add(start_debugger, password)
+			GLib.idle_add(start_debugger, password)
 		dialog.destroy()
+		return False
 	
+	def enable_debugpy(self, action, parameter, shell):
+		if have_debugpy is False:
+			self.missing_python_module("debugpy", "debugpy")
+			self.debugpy_action.set_enabled(False)
+			return False
+		else:
+			try:
+				host, port = debugpy.listen(DEBUG_PORT)
+				msgtype = Gtk.MessageType.INFO
+				msg = _("Rhythmbox is now listening for Debug Adapter Protocol connections on port %d. You can now attach to it using a compatible debugger such as vimspector, nvim-dap or Visual Studio Code.") % port
+				self.debugpy_action.set_enabled(False)
+			except Exception as e:
+				msgtype = Gtk.MessageType.ERROR
+				msg = _("Unable to start Debug Adapter Protocol listener: %s") % str(e)
+
+		self.message_dialog(msg, msgtype)
+		return False
+
+	def missing_python_module(self, module, debugger):
+		msg = _("The %s Python module is not available. Install the module and then restart Rhythmbox to enable debugging with %s.") % (module, debugger)
+		self.message_dialog(msg, Gtk.MessageType.ERROR)
+
+	def message_dialog(self, msg, msgtype=Gtk.MessageType.INFO):
+		dialog = Gtk.MessageDialog(None, 0, msgtype, Gtk.ButtonsType.OK, msg)
+		dialog.connect("response", self.message_dialog_response)
+		dialog.show()
+
+
+	def message_dialog_response(self, dialog, rsp):
+		dialog.destroy()
+
 	def destroy_console(self, *args):
 		self.window.destroy()
 		self.window = None
@@ -172,7 +211,7 @@ class PythonConsole(Gtk.ScrolledWindow):
 		Gtk.ScrolledWindow.__init__(self)
 
 		self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
-		self.set_shadow_type(Gtk.ShadowType.IN)
+		self.set_shadow_type(Gtk.ShadowType.NONE)
 		self.view = Gtk.TextView()
 		self.view.modify_font(Pango.font_description_from_string('Monospace'))
 		self.view.set_editable(True)
@@ -208,12 +247,12 @@ class PythonConsole(Gtk.ScrolledWindow):
 		# Set up hooks for standard output.
 		self.stdout = gtkoutfile(self, sys.stdout.fileno(), self.normal)
 		self.stderr = gtkoutfile(self, sys.stderr.fileno(), self.error)
-		
+
 		# Signals
 		self.view.connect("key-press-event", self.__key_press_event_cb)
 		buffer.connect("mark-set", self.__mark_set_cb)
-		
- 		
+
+
 	def __key_press_event_cb(self, view, event):
 		if event.keyval == Gdk.KEY_D and \
 		   event.state == Gdk.ModifierType.CONTROL_MASK:
@@ -243,7 +282,7 @@ class PythonConsole(Gtk.ScrolledWindow):
 				cur = self.get_end_iter()
 				
 			buffer.place_cursor(cur)
-			GObject.idle_add(self.scroll_to_end)
+			GLib.idle_add(self.scroll_to_end)
 			return True
 		
 		elif event.keyval == Gdk.KEY_Return:
@@ -287,7 +326,7 @@ class PythonConsole(Gtk.ScrolledWindow):
 			cur = self.get_end_iter()
 			buffer.move_mark(inp_mark, cur)
 			buffer.place_cursor(cur)
-			GObject.idle_add(self.scroll_to_end)
+			GLib.idle_add(self.scroll_to_end)
 			return True
 
 		elif event.keyval == Gdk.KEY_KP_Down or \
@@ -295,7 +334,7 @@ class PythonConsole(Gtk.ScrolledWindow):
 			# Next entry from history
 			view.emit_stop_by_name("key_press_event")
 			self.history_down()
-			GObject.idle_add(self.scroll_to_end)
+			GLib.idle_add(self.scroll_to_end)
 			return True
 
 		elif event.keyval == Gdk.KEY_KP_Up or \
@@ -303,7 +342,7 @@ class PythonConsole(Gtk.ScrolledWindow):
 			# Previous entry from history
 			view.emit_stop_by_name("key_press_event")
 			self.history_up()
-			GObject.idle_add(self.scroll_to_end)
+			GLib.idle_add(self.scroll_to_end)
 			return True
 
 		elif event.keyval == Gdk.KEY_KP_Left or \
@@ -382,23 +421,23 @@ class PythonConsole(Gtk.ScrolledWindow):
 			start_iter = Gtk.TextIter()
 			start_iter = buffer.get_iter_at_offset(offset)
 			buffer.apply_tag(tag, start_iter, self.get_end_iter())
-		GObject.idle_add(self.scroll_to_end)
- 	
- 	def eval(self, command, display_command = False):
+		GLib.idle_add(self.scroll_to_end)
+
+	def eval(self, command, display_command = False):
 		buffer = self.view.get_buffer()
 		lin = buffer.get_mark("input-line")
 		buffer.delete(self.get_iter_at_mark(lin),
 			      self.get_end_iter())
- 
+
 		if isinstance(command, list) or isinstance(command, tuple):
- 			for c in command:
-		 		if display_command:
-		 			self.write(">>> " + c + "\n", self.command)
- 				self.__run(c)
+			for c in command:
+				if display_command:
+					self.write(">>> " + c + "\n", self.command)
+				self.__run(c)
 		else:
-	 		if display_command:
-	 			self.write(">>> " + c + "\n", self.command)
-			self.__run(command) 
+			if display_command:
+				self.write(">>> " + c + "\n", self.command)
+			self.__run(command)
 
 		cur = self.get_end_iter()
 		buffer.move_mark_by_name("input-line", cur)
@@ -407,7 +446,7 @@ class PythonConsole(Gtk.ScrolledWindow):
 		buffer.move_mark_by_name("input", cur)
 		self.view.scroll_to_iter(self.get_end_iter(), 0.0, False, 0.5, 0.5)
 	
- 	def __run(self, command):
+	def __run(self, command):
 		sys.stdout, self.stdout = self.stdout, sys.stdout
 		sys.stderr, self.stderr = self.stderr, sys.stderr
 		
@@ -415,9 +454,9 @@ class PythonConsole(Gtk.ScrolledWindow):
 			try:
 				r = eval(command, self.namespace, self.namespace)
 				if r is not None:
-					print `r`
+					print(repr(r))
 			except SyntaxError:
-				exec command in self.namespace
+				exec(command, self.namespace)
 		except:
 			if hasattr(sys, 'last_type') and sys.last_type == SystemExit:
 				self.destroy()
@@ -448,8 +487,8 @@ class gtkoutfile:
 	def readlines(self):     return []
 	def write(self, s):      self.console.write(s, self.tag)
 	def writelines(self, l): self.console.write(l, self.tag)
-	def seek(self, a):       raise IOError, (29, 'Illegal seek')
-	def tell(self):          raise IOError, (29, 'Illegal seek')
+	def seek(self, a):       raise IOError((29, 'Illegal seek'))
+	def tell(self):          raise IOError((29, 'Illegal seek'))
 	truncate = tell
 
 # ex:noet:ts=8:
